# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Placeholder for centralized package logging and version control."""

import logging


__version__ = '0.4.6'


class CustomFormatter(logging.Formatter):
    """
    Custom logging formatter with colors.
    """
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: f'[{grey}%(levelname).1s{reset}] %(message)s',
        logging.INFO: f'[{grey}%(levelname).1s{reset}] %(message)s',
        logging.WARNING: f'[{yellow}%(levelname).1s{reset}] %(message)s',
        logging.ERROR: f'[{red}%(levelname).1s{reset}] %(message)s',
        logging.CRITICAL: f'[{bold_red}%(levelname).1s{reset}] %(message)s'
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        return logging.Formatter(log_fmt).format(record)


log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

ch.setFormatter(CustomFormatter())
# add the handlers to the logger
log.addHandler(ch)


def add_file_handler():
    """
    Create and add a FileHandler to log debug messages into a file.
    """
    # create file handler which logs even debug messages
    fhandler = logging.FileHandler('.prune-merged.log')
    fhandler.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    fhandler.setFormatter(
        logging.Formatter('%(levelname)s - %(message)s (%(filename)s:%(lineno)d)'))
    log.addHandler(fhandler)



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Offers utility functions to obtain and log local and remote branches
that can be potentially pruned.
"""

import getpass
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from dateutil.relativedelta import relativedelta
import yaml
from git import Repo
from gitdb.exc import BadName
from github import Github

from . import log


class Branches:
    # pylint: disable=too-few-public-methods
    """
    Contains generic information in order to query merged remote branches.
    """

    def __init__(self, config: Path, months: int) -> None:
        self._config: Path = config
        self._months: int = months
        self._repo: str = ''


class GithubBranches(Branches):
    # pylint: disable=too-few-public-methods
    """
    Github specialization for the Branches class.
    """

    def __init__(self, config: Path, months: int) -> None:
        super().__init__(config, months)
        self._gh: Optional[Github] = None
        self._create_github()

    def _create_github(self) -> Github:
        with open(self._config, 'r', encoding='utf-8') as conf_file:
            loaded_conf = yaml.safe_load(conf_file)

        self._repo = loaded_conf.get('repo')
        if token := loaded_conf.get('token'):
            self._gh = Github(token)  # pylint: disable=invalid-name
        else:
            username = loaded_conf.get('user')
            assert username is not None
            pword = getpass.getpass(prompt=f"{username}'s password: ")
            self._gh = Github(username, pword)  # pylint: disable=invalid-name
            del pword

    @property
    def branches(self) -> List['github.PullRequest']:
        """
        Search for all PR issues that pertain to the current user, repository,
        that have been closed and merged.
        """
        assert self._gh is not None
        # Get PRs (issues of type PR) that pertain to the GitHub user, for this
        # specific repo. Select those that are closed and merged.
        qualifiers = {
            'state': 'closed',
            'is': 'merged',
            'author': self._gh.get_user().login,
            'repo': self._repo,
            'type': 'pr'
        }
        if self._months:
            months_old = datetime.now() - relativedelta(months=self._months)
            qualifiers['closed'] = f'>{months_old.strftime("%Y-%m-%d")}'

        return [pr.as_pull_request() for pr in self._gh.search_issues('', **qualifiers)]


def log_branch_list(branch_loc: str, branches: List[str]) -> None:
    """
    Log string elements in a list that represent branch names.
    Locations stands for private or remote, that will be logged in the output.
    """
    if len(branches) > 0:
        log.info(
            'There are %s %s branches that can be pruned:',
            len(branches),
            branch_loc)
        for branch_name in branches:
            log.info("\t%s", branch_name)
    else:
        log.info('No %s branches to prune.', branch_loc)


def list_branches(config: Path, months: int) -> None:
    """
    Obtain and log all local and remote branches that are older than months.
    """
    gh_merged = GithubBranches(config, months).branches
    repo = Repo()
    # Loop through local head commits and compare its SHA against GitHub head SHA.
    local_branches = []
    for head in repo.heads:
        if any(head.commit.hexsha == gh_pr.head.sha for gh_pr in gh_merged):
            local_branches.append(head.name)
    # Log local branches found.
    log_branch_list("local", local_branches)

    # Loop through GitHub PRs and rev_parse the remote branch name to be able
    # to compare the SHAs.
    remote_branches = []
    for gh_pr in gh_merged:
        remote_name = f'origin/{gh_pr.head.ref}'
        try:
            sha = repo.rev_parse(remote_name)
        except BadName:
            continue

        if str(sha) == gh_pr.head.sha:
            remote_branches.append(remote_name)
    # Log remote branches found.
    log_branch_list("remote", remote_branches)



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Main aplication CLI.
"""

import sys
from pathlib import Path
from typing import Optional

import click

from . import __version__, add_file_handler, log
from .config import DEFAULT_CFG_PATH, create_config
from .exceptions import exception_handler
from .list import list_branches, GithubBranches
from .prune import prune_local, prune_remote

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def print_version(ctx, param, value):
    # pylint: disable=unused-argument
    """
    Print version and exit.
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo(f'git-prune-merged version: {__version__}')
    ctx.exit()


@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True)
@click.pass_context
@exception_handler
@click.option('--debug', '-d', is_flag=True,
              help='Log debug messages into a file.', default=False)
@click.option('--version', '-v', is_flag=True, callback=print_version,
              expose_value=False, is_eager=True, help='Print version and exit.')
@click.option('--config', '-c', 'configuration', type=click.Path(dir_okay=False, exists=True),
              default=DEFAULT_CFG_PATH, help='Path to the config file.')
@click.option('--local/--remote', '-l/-r', is_flag=True, default=True, show_default=True,
              help='Select to prune only local or remote merged branches.')
@click.option('--all', 'all_branches', is_flag=True, default=False,
              help='Prune both remote and local merged branches.')
@click.option('--months', type=int, default=0,
              help='Select PRs older than specified months.')
@click.option('--yes', is_flag=True, default=False, help='Do not ask for confirmation.')
@click.option('--dry-run', is_flag=True, default=False,
              help='Simulated run, do not delete branches.')
def cli(ctx, debug: bool, configuration: Path, local: bool,
        all_branches: bool, months: int, yes: bool, dry_run: bool) -> None:
    # pylint: disable=too-many-arguments, unused-argument
    """
    Prune local and remote branches that have been merged, even if it
    has been merged by rebasing.

    Currently supported only for GitHub projects.

    Branches are deleted based on the SHA1 of the HEAD of the branch
    and the status of the Pull Request in GitHub.
    """
    if debug:
        add_file_handler()
    if ctx.invoked_subcommand is None:
        # With no subcommmand, we prune the branches
        gh_pr = GithubBranches(configuration, months).branches
        if local or all_branches:
            prune_local(yes, dry_run, gh_pr)
        if not local or all_branches:
            prune_remote(yes, dry_run, gh_pr)


@cli.command()
@exception_handler
@click.option('--path', '-p', type=click.Path(dir_okay=False),
              default=DEFAULT_CFG_PATH, help='Path to the config file.')
@click.option('--token', '-t', type=str,
              help='Specify token for GitHub login [recommended].')
@click.option('--user', '-u', type=str,
              help='Specify user for GitHub login.')
@click.option('--repo', '-r', type=str, required=True)
def config(
        path: Path,
        token: Optional[str],
        user: Optional[str],
        repo: Optional[str]) -> None:
    """
    Setup the configuration file to use.
    """
    if not token and not user:
        log.error('Need at least the token or the user to login.')
        sys.exit(1)
    create_config(path, token, user, repo)


@cli.command()
@exception_handler
@click.option('--config', '-c', 'configuration', type=click.Path(dir_okay=False, exists=True),
              default=DEFAULT_CFG_PATH, help='Path to the config file.')
@click.option('--months', type=int, default=0,
              help='Select PRs older than specified months.')
def ls(configuration: Path, months: int) -> None:
    # pylint: disable=invalid-name
    """
    List branches that will be deleted in a prune.
    """
    list_branches(configuration, months)



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Centralized exception handling for the entire package.
"""


from functools import wraps

from github.GithubException import GithubException

from . import log


def exception_handler(func):
    """
    Can be used as a decorator for CLI functions to handle all posible
    exceptions.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except GithubException as exc:
            log.critical('Failed to connect to GitHub: %s', exc)

    return wrapper



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Main entry point to the prune-rebase CLI."""

from .main import cli

if __name__ == '__main__':
    cli()  # pylint: disable=no-value-for-parameter



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Handle the actual prune of all local and remote branches.
"""

from typing import List

from git import Reference, Repo
from gitdb.exc import BadName

from . import log


def _prompt_confirmation(branch_name: str) -> bool:
    while True:
        log.info("Going to remove %s", branch_name)
        answer = input("    Are you sure? [Y/n] ")
        if not answer or any(char in answer for char in ['y', 'Y']):
            return True
        if any(char in answer for char in ['n', 'N']):
            return False


def prune_remote(yes: bool, dry_run: bool, gh_pr: List['github.PullRequest']):
    """
    Prune remote branches based on its SHA.
    """
    deleted = 0
    repo = Repo()
    for preq in gh_pr:
        remote_name = f'origin/{preq.head.ref}'
        try:
            sha = repo.rev_parse(remote_name)
        except BadName:
            continue

        if str(sha) == preq.head.sha:
            answer = True
            if not yes:
                answer = _prompt_confirmation(remote_name)
            if answer:
                log.debug("Removing %s...", remote_name)
                if not dry_run:
                    repo.remote().push(refspec=(f":{preq.head.ref}"))
                deleted += 1
                log.info("[deleted] .... %s", remote_name)
            else:
                log.info("Skipping %s", remote_name)

    log.info("All (%s) remote branches pruned.", deleted)


def prune_local(yes: bool, dry_run: bool, gh_pr: List['github.PullRequest']):
    """
    Prune local branches based on its SHA.
    """
    deleted = 0
    repo = Repo()
    for head in repo.heads:
        if any(head.commit.hexsha == preq.head.sha for preq in gh_pr):
            answer = True
            if not yes:
                answer = _prompt_confirmation(head.name)
            if answer:
                try:
                    if not dry_run:
                        Reference.delete(repo, head.path)
                    deleted += 1
                    log.info("[deleted] .... %s", head.name)
                except OSError:
                    log.error("%s", head, exc_info=True)
            else:
                log.info("Skipping %s", head.name)

    log.info("All (%s) local branches pruned.", deleted)



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""git-prune-merged package setup."""

from setuptools import setup, find_packages
import src

with open("README.md", "r", encoding="utf8") as readme_file:
    readme = readme_file.read()

requirements = ["Click", "gitpython>3.1", "pygithub>=1.55", "pyyaml>=6.0", "python-dateutil"]

setup(
    name="git-prune-merged",
    version=src.__version__,
    author="Albert Esteve",
    author_email="aesteve@redhat.com",
    description="A git plugin to prune local and remote branches",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/aesteve-rh/git-prune-merged",
    packages=find_packages(),
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'git-prune-merged=src.main:cli',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)



# SPDX-FileCopyrightText: 2022 Albert Esteve <aesteve@redhat.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Configuration file stored at project's root .git folder.
The file stores information required to interact with GH Rest API.
"""

from pathlib import Path
from typing import Optional

import yaml

from . import log

DEFAULT_CFG_PATH = Path(Path.cwd(), '.git/.gitpruneconf')


def create_config(
        path: Path,
        token: Optional[str],
        user: Optional[str],
        repo: Optional[str]) -> None:
    # pylint: disable=unused-argument
    """
    Create configuration file for GitHub.
    """
    config = {key: value for key, value in locals().items()
              if value and key != 'path'}
    if path.is_file():
        log.warning("A configuration file already exists at '%s'", path)
        log.info("If you continue the configuration will be overwritten.")
        while True:
            answer = input("    Are you sure? [Y/n] ")
            if not answer or any(char in answer for char in ['y', 'Y']):
                break
            if any(char in answer for char in ['n', 'N']):
                return

    with open(path, 'w', encoding='utf-8') as conf_file:
        yaml.dump(config, conf_file)
